// gpd2svg.cpp		ver.0.01
//copyright 2005 Beer[Tsutsumi Hiroyuki]

#include	<Be.h>

#define	VER		"ver.0.01"
#define	DATE	"2005. 2.11"

enum{				//object-type
	PICK=0,
	POINT=0x10,
	LINE=0x20,
	CURVE,
	RECT=0x30,
	ELLIPSE=0x40,
	POLY=0x50,
	SHAPE,
	TEXT=0x80,
	BITMAP=0x90
};

void	avcol(uchar *fc,uchar *bc,uchar *g,uint32 *c,float *a);
void	quit(char *s);

off_t	rbl=0;
char	*rb=NULL;
off_t	wbl=0x100000;
char	*wb=NULL;

int	main(int argc,char **argv)
{
	if(argc!=2)quit("gpd2svg ver.0.01\nusage: gpd2svg gprod-file");

	BFile	rf(argv[1],B_READ_ONLY);
	if(!rf.IsReadable())quit("cannot open gprod-file");
	rf.GetSize(&rbl);
	rb=new char[rbl];
	rf.Read(rb,rbl);
	char	*rp=rb;
	wb=new char[wbl];
	char	*wp=wb;

	//read header
	char	*hd=rp;
	rp+=16;
	if(strncmp(hd,"gprod_01",8))quit("err: is not gprod-file");
	int	cw=B_BENDIAN_TO_HOST_INT16(*((uint16*)(hd+8)));
	int	ch=B_BENDIAN_TO_HOST_INT16(*((uint16*)(hd+10)));
	//read color-def
	uchar	cc=*(rp+5);
	char	*cd=rp+8;
	uint32	col[cc];
	for(int i=0;i<cc;i++)col[i]=B_BENDIAN_TO_HOST_INT32(*(uint32*)(cd+i*4));
	rp+=8+B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+6)));
	//read pattern-def
	uchar	pc=*(rp+5);
	char	*pd=rp+8;
	rp+=8+B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+6)));
	//read font-name
	uchar	fc=*(rp+5);
	char	*fd=rp+8;
	rp+=8+B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+6)));
	//write svg header
	wp+=sprintf(wp,"<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n");
	wp+=sprintf(wp,"<!DOCTYPE svg>\n");
	wp+=sprintf(wp,"<svg width=\"%d\" height=\"%d\" viewBox=\"0 0 %d %d\" overflow=\"visible\" enable-background=\"new 0 0 %d %d\" xml:space=\"preserve\"",cw,ch,cw,ch,cw,ch);
	wp+=sprintf(wp," stroke-linecap=\"round\" stroke-linejoin=\"round\">\n");
	wp+=sprintf(wp,"<g>\n");
	//read objects
	uint16	objs=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+4)));
	char	*dt=rp+8;
	rp+=8;
	char	tb[0x4000];
	for(int ii=0;ii<objs;ii++){
		uchar	type=*(rp++);
		char	*tp=tb;
		int	l,t,r,b,fc,bc,fg,bg,pen;
		switch(type){
			case RECT:{
				l=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+0)));
				t=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+2)));
				r=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+4)));
				b=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+6)));
				fc=*((uchar*)(rp+8));
				bc=*((uchar*)(rp+9));
				fg=*((uchar*)(rp+10));
				bg=*((uchar*)(rp+11));
				pen=*((uchar*)(rp+12));
				rp+=13;
	tp+=sprintf(tp,"rect x=\"%5.1f\" y=\"%5.1f\" width=\"%5.1f\" height=\"%5.1f\"",(float)l,(float)t,(float)r-l+1,(float)b-t+1);
	tp+=sprintf(tp," stroke-width=\"%d\"",pen);
	tp+=sprintf(tp," stroke-linecap=\"square\" stroke-linejoin=\"miter\"");
 			}break;
			case ELLIPSE:{
				l=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+0)));
				t=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+2)));
				r=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+4)));
				b=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+6)));
				fc=*((uchar*)(rp+8));
				bc=*((uchar*)(rp+9));
				fg=*((uchar*)(rp+10));
				bg=*((uchar*)(rp+11));
				pen=*((uchar*)(rp+12));
				rp+=13;
	tp+=sprintf(tp,"ellipse cx=\"%5.1f\" cy=\"%5.1f\" rx=\"%5.1f\" ry=\"%5.1f\"",(float)(l+r)/2,(float)(t+b)/2,(float)(r-l+1)/2,(float)(b-t+1)/2);
	tp+=sprintf(tp," stroke-width=\"%d\"",pen);
			}break;
			case TEXT:{
				l=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+0)));
				t=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+2)));
				r=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+4)));
				b=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+6)));
				fc=*((uchar*)(rp+8));
				bc=*((uchar*)(rp+9));
				fg=*((uchar*)(rp+10));
				bg=*((uchar*)(rp+11));
				int	fon=*((uchar*)(rp+12));
				int	fos=*((uchar*)(rp+13));
				int	tl=*((uchar*)(rp+14));
				rp+=15+tl;
			}break;
			case BITMAP:{
				l=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+0)));
				t=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+2)));
				r=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+4)));
				b=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+6)));
				int	bl=((r-l+1)*(b-t+1))*4;
				rp+=8+bl;
			}break;
			case LINE:{
				int	pts=*((uchar*)(rp+0));
				BPoint	pt[pts];
				for(int i=0;i<pts;i++){
					pt[i].x=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+1+i*4)));
					pt[i].y=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+3+i*4)));
				}
				fc=*((uchar*)(rp+pts*4+1));
				bc=*((uchar*)(rp+pts*4+2));
				fg=*((uchar*)(rp+pts*4+3));
				pen=*((uchar*)(rp+pts*4+4));
				rp+=5+pts*4;
	tp+=sprintf(tp,"path d=\"M%5.1f,%5.1f",pt[0].x,pt[0].y);
				for(int i=1;i<pts;i++){
	tp+=sprintf(tp," L%5.1f,%5.1f",pt[i].x,pt[i].y);
				}
	tp+=sprintf(tp,"\"");
	tp+=sprintf(tp," stroke-width=\"%d\"",pen);
			}break;
			case POLY:{
				int	pts=*((uchar*)(rp+0));
				BPoint	pt[pts];
				for(int i=0;i<pts;i++){
					pt[i].x=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+1+i*4)));
					pt[i].y=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+3+i*4)));
				}
				fc=*((uchar*)(rp+pts*4+1));
				bc=*((uchar*)(rp+pts*4+2));
				fg=*((uchar*)(rp+pts*4+3));
				bg=*((uchar*)(rp+pts*4+4));
				pen=*((uchar*)(rp+pts*4+5));
				rp+=6+pts*4;
	tp+=sprintf(tp,"path d=\"M%5.1f,%5.1f",pt[pts-1].x,pt[pts-1].y);
				for(int i=0;i<pts;i++){
	tp+=sprintf(tp," L%5.1f,%5.1f",pt[i].x,pt[i].y);
				}
	tp+=sprintf(tp,"\"");
	tp+=sprintf(tp," stroke-width=\"%d\"",pen);
			}break;
			case CURVE:{
				int	pts=*((uchar*)(rp+0));
				BPoint	pt[pts];
				for(int i=0;i<pts;i++){
					pt[i].x=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+1+i*4)));
					pt[i].y=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+3+i*4)));
				}
				fc=*((uchar*)(rp+pts*4+1));
				bc=*((uchar*)(rp+pts*4+2));
				fg=*((uchar*)(rp+pts*4+3));
				pen=*((uchar*)(rp+pts*4+4));
				rp+=5+pts*4;
				BPoint	q[4];
	tp+=sprintf(tp,"path d=\"M%5.1f,%5.1f",pt[0].x,pt[0].y);
				for(int i=0;i<pts-2;i++){
					q[0]=(i==0)?pt[i]:BPoint((pt[i].x+pt[i+1].x)/2,(pt[i].y+pt[i+1].y)/2);
					q[3]=(i==pts-3)?pt[i+2]:BPoint((pt[i+1].x+pt[i+2].x)/2,(pt[i+1].y+pt[i+2].y)/2);
					q[1]=BPoint((q[0].x+pt[i+1].x*2)/3,(q[0].y+pt[i+1].y*2)/3);
					q[2]=BPoint((q[3].x+pt[i+1].x*2)/3,(q[3].y+pt[i+1].y*2)/3);
	tp+=sprintf(tp," C%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f",q[1].x,q[1].y,q[2].x,q[2].y,q[3].x,q[3].y);
				}
	tp+=sprintf(tp,"\"");
	tp+=sprintf(tp," stroke-width=\"%d\"",pen);
			}break;
			case SHAPE:{
				int	pts=*((uchar*)(rp+0));
				BPoint	pt[pts];
				for(int i=0;i<pts;i++){
					pt[i].x=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+1+i*4)));
					pt[i].y=B_BENDIAN_TO_HOST_INT16(*((uint16*)(rp+3+i*4)));
				}
				fc=*((uchar*)(rp+pts*4+1));
				bc=*((uchar*)(rp+pts*4+2));
				fg=*((uchar*)(rp+pts*4+3));
				bg=*((uchar*)(rp+pts*4+4));
				pen=*((uchar*)(rp+pts*4+5));
				rp+=6+pts*4;
				BPoint	q[4];
				q[0]=BPoint((pt[0].x+pt[1].x)/2,(pt[0].y+pt[1].y)/2);
	tp+=sprintf(tp,"path d=\"M%5.1f,%5.1f",q[0].x,q[0].y);
				for(int i=0;i<pts;i++){
					q[0]=BPoint((pt[i].x+pt[(i+1)%pts].x)/2,(pt[i].y+pt[(i+1)%pts].y)/2);
					q[3]=BPoint((pt[(i+1)%pts].x+pt[(i+2)%pts].x)/2,(pt[(i+1)%pts].y+pt[(i+2)%pts].y)/2);
					q[1]=BPoint((q[0].x+pt[(i+1)%pts].x*2)/3,(q[0].y+pt[(i+1)%pts].y*2)/3);
					q[2]=BPoint((q[3].x+pt[(i+1)%pts].x*2)/3,(q[3].y+pt[(i+1)%pts].y*2)/3);
	tp+=sprintf(tp," C%5.1f,%5.1f,%5.1f,%5.1f,%5.1f,%5.1f",q[1].x,q[1].y,q[2].x,q[2].y,q[3].x,q[3].y);
				}
	tp+=sprintf(tp,"\"");
	tp+=sprintf(tp," stroke-width=\"%d\"",pen);
			}break;
			default:{
			}break;
		}
		if((type==RECT)||(type==ELLIPSE)||(type==LINE)||(type==POLY)||(type==CURVE)||(type==SHAPE)){
			uint32	c1,c2;
			float	a1,a2;
			avcol((uchar*)cd+fc*4,(uchar*)cd+bc*4,(uchar*)pd+fg*32,&c1,&a1);
			if((type==RECT)||(type==ELLIPSE)||(type==POLY)||(type==SHAPE)){
				avcol((uchar*)cd+fc*4,(uchar*)cd+bc*4,(uchar*)pd+bg*32,&c2,&a2);
			}else{
				c2=0;a2=0;
			}

//printf("c1=%06x,c2=%06x,a1=%4.2f,a2=%4.2f\n",c1,c2,a1,a2);
			if((a1!=0xff)&&(a2==0xff)){
	wp+=sprintf(wp,"<%s stroke=\"none\" fill=\"#%06x\"/>\n",tb,c2);
			}
			if((a1!=a2)&&(a2!=0x00)&&(a2!=0xff)){
	wp+=sprintf(wp,"<%s stroke=\"none\" fill=\"#%06x\" opacity=\"%4.2f\"/>\n",tb,c2,a2);
			}
			if((a1==0xff)&&(a2!=0xff)){
	wp+=sprintf(wp,"<%s stroke=\"#%06x\" fill=\"none\"/>\n",tb,c1);
			}
			if((a1!=a2)&&(a1!=0x00)&&(a1!=0xff)){
	wp+=sprintf(wp,"<%s stroke=\"#%06x\" fill=\"none\" opacity=\"%4.2f\"/>\n",tb,c1,a1);
			}
			if((a1==0xff)&&(a2==0xff)){
	wp+=sprintf(wp,"<%s stroke=\"#%06x\" fill=\"#%06x\"/>\n",tb,c1,c2);
			}
			if((a1==a2)&&(a1!=0x00)&&(a1!=0xff)){
	wp+=sprintf(wp,"<%s stroke=\"#%06x\" fill=\"#%06x\" opacity=\"%4.2f\"/>\n",tb,c1,c2,a1);
			}

		}
	}
	wp+=sprintf(wp,"</g>\n");
	wp+=sprintf(wp,"</svg>\n");

	char	s[256];
	sprintf(s,"%s.svg",argv[1]);
	BFile	wf(s,B_WRITE_ONLY|B_CREATE_FILE|B_ERASE_FILE);
	if(!wf.IsWritable())quit("cannot open svg-file");
	wf.Write(wb,wp-wb);

	quit("");
}

void avcol(uchar *fc,uchar *bc,uchar *g,uint32 *c,float *a)
{
	uchar	fa,fr,fg,fb,ba,br,bg,bb;
	fa=*(fc++);	fr=*(fc++);	fg=*(fc++);	fb=*(fc++);
	ba=*(bc++);	br=*(bc++);	bg=*(bc++);	bb=*(bc++);
	int	p=0;
	for(int i=0;i<32;i++)for(uchar b=*(g++);b;b>>=1)if(b&1)p++;
	float	fk,bk;
	fk=(float)p/256;	bk=(float)(256-p)/256;
	*c=((uchar)(fk*fr+bk*br)<<16)|((uchar)(fk*fg+bk*bg)<<8)|((uchar)(fk*fb+bk*bb));
	*a=(fk*fa+bk*ba)/255;
}

void quit(char *s)
{
	if(rb)delete rb;
	if(wb)delete wb;
	if(*s)printf("%s\n",s);
	exit(*s);
}
